;;; ABSTRACT-SYNTAX.SCM
;;;
;;; Abstract syntax for INTEX+BIND programs and expressions

;;; ***************************************************************
;;; Constants

(define binop-names '(+ - * div mod))

(define binop-name? (lambda (x) (memq x binop-names)))

(define keywords (append '(program bind) binop-names))

(define keyword? (lambda (x) (memq x keywords)))


;;; ***************************************************************
;;; Programs

(define make-program
  (lambda (formals body)
    (if (not (forall? symbol? formals))
      (error "MAKE-PROGRAM: first argument should be a list of symbols -- "
             formals)
      (list 'program formals body))))

(define program? 
  (lambda (sexp)
    (and (not (null? sexp))
         (eq? (car sexp) 'program)
         ;; The following ferrets out malformed programs
         (or (and (list? sexp)
                  (= (length sexp) 3)
                  (forall? symbol? (program-formals sexp)))
             (error "Malformed program" sexp)))))

(define program-formals second)
(define program-body third)

;;; ***************************************************************
;;; Expressions

;;; ---------------------------------------------------------------
;;; Literals: 
;;; Represent INTEX+BIND integer literals as Scheme integers.

(define make-literal (lambda (lit) lit))

;; Extract the value from an LITERAL node. In this representation,
;; this is the identity. 
(define literal-value (lambda (lit) lit))

(define literal? integer?)

;;; ---------------------------------------------------------------
;;; Variable references: 
;;; Represent these as Scheme symbols (note: this is OK since
;;; INTEX spec says variable names aren't case sensitive.)

(define make-varref (lambda (sym) sym))

;; Extract the variable name from a VARREF node. In this representation,
;; this is the identity. 
(define varref-name (lambda (sym) sym))

(define varref? 
  (lambda (sexp) 
    (and (symbol? sexp)
         ;; Keywords and operators are not variable references
         (not (keyword? sexp)))))

;;; ---------------------------------------------------------------
;;; Binary operator applications

(define make-binapp
  (lambda (rator rand1 rand2)
    (if (not (binop-name? rator))
      (error "MAKE-BINAPP: first argument should be a binary operator name -- "
             rator)
      (list rator rand1 rand2))))

(define binapp-rator first)
(define binapp-rand1 second)
(define binapp-rand2 third)

(define binapp? 
  (lambda (sexp)
    (and (not (null? sexp))
         (binop-name? (car sexp))
         ;; The following ferrets out malformed binops
         (or (and (list? sexp)
                  (= (length sexp) 3))
             (error "Malformed binapp" sexp)))))
         
;;; ---------------------------------------------------------------
;;; Local binding

(define make-bind
  (lambda (name defn body)
    (if (not (symbol? name))
      (error "MAKE-BIND: first argument should be a symbol --" name)
      (list 'bind name defn body))))

(define bind-name second)
(define bind-defn third)
(define bind-body fourth)

(define bind? 
  (lambda (sexp)
    (and (not (null? sexp))
         (eq? (car sexp) 'bind)
         ;; The following ferrets out malformed binops
         (or (and (list? sexp)
                  (= (length sexp) 4)
                  (symbol? (bind-name sexp)))
             (error "Malformed bind" sexp)))))

(define make-bind
  (lambda (name defn body)
    (if (not (symbol? name))
      (error "MAKE-BIND: first argument should be a symbol --" name)
      (list 'bind name defn body))))

(define bind-name second)
(define bind-defn third)
(define bind-body fourth)

