;; RENAME.SCM for INTEX+BIND programs.

;; RENAME, applied to EXP and a set of "forbidden" names
;; RESERVED-VAR-NAMES, first tests whether (FREE-VARS EXP)
;; is a subset (not necessarily proper) of RESERVED-VAR-NAMES,
;; and if this test succeeds, it returns a new expression
;; EXP' which is identical to EXP except that the name of
;; every binding variable occurrence in EXP is changed to
;; a fresh variable name not in RESERVED-VAR-NAMES.

;; RENAME uses auxiliary functions RENAME-FREE and NAME-NOT-IN, which 
;; are both defined in this file, as well as several help functions
;; in the file (1) "set-list-nodups.scm":
;;      set-subset?, set-union, set-singleton,
;; in the file (2) "abstract-syntax.scm":
;;      literal?, varref?, binapp?, make-binapp, binapp-rator,
;;      binapp-rand1, binapp-rand2, bind?, bind-name, bind-defn,
;;      bind-body, make-bind,
;; in the file (3) "free-binding-bound-vars.scm":
;;      free-vars, binding-vars

(define rename
  (lambda (exp reserved-var-names)
    (cond ((not (set-subset? (free-vars exp) reserved-var-names))
	   (error "there are free variable occurrences not in -- "
		  reserved-var-names))
	  ((literal? exp) exp)
	  ((varref? exp) exp)
	  ((binapp? exp)
	   (make-binapp (binapp-rator exp)
			(rename (binapp-rand1 exp) reserved-var-names)
			(rename (binapp-rand2 exp) reserved-var-names)))
	  ((bind? exp)
	   (let* ((old-name (bind-name exp))       ;; save old bind-name
		  (old-defn (bind-defn exp))       ;; save old bind-defn
		  (old-body (bind-body exp))       ;; save old bind-body
                  (new-defn (rename old-defn       ;; define new bind-defn
			        reserved-var-names))
		  (new-body (rename old-body       ;; define new bind-body
				(set-union (set-singleton old-name)
					   reserved-var-names)))
		  (new-name (name-not-in old-name  ;; define new bind-name
				(set-union (binding-vars new-body)
					   reserved-var-names))))
	     (make-bind new-name
			new-defn
			(rename-free old-name new-name new-body))))
	  (else (error "unknown expression type --" exp)))))

;; RENAME-FREE replaces all free occurrences of
;; OLD in EXP by NEW, provided there are no binding
;; ocurrences of NEW in EXP. For example,
;;
;; > (rename-free 'a 'b '(* a 3))
;; (* b 3)
;;
;; > (rename-free 'a 'b '(* a (bind a (+ a 1) (* a 2))))  
;; (* b (bind a (+ b 1) (* a 2)))
;; 
;; > (rename-free 'a 'b '(* a (bind b (+ a 1) (* b 2)))) 
;; ; there are binding occurrences of -- b
;;
;; > (rename-free 'a 'b '(bind c (+ a 1) (+ c c)))          
;; (bind c (+ b 1) (+ c c))

(define rename-free
  (lambda (old new exp)
    (cond ((set-member? new (binding-vars exp))
	   (error "there are binding occurrences of  --" new))
	  ((literal? exp) exp)
	  ((varref? exp)
	   (if (eq? old exp) new exp))
	  ((binapp? exp)
	   (make-binapp (binapp-rator exp)
			(rename-free old new (binapp-rand1 exp))
			(rename-free old new (binapp-rand2 exp))))
	  ((bind? exp)
	   (if (eq? old (bind-name exp)) 
	       (make-bind (bind-name exp)
			  (rename-free old new (bind-defn exp))
			  (bind-body exp))
	       (make-bind (bind-name exp)
			  (rename-free old new (bind-defn exp))
			  (rename-free old new (bind-body exp)))))
	  (else (error "unknown expression type --" exp)))))

;; NAME-NOT-IN returns the first version of the "root" of the symbol 
;; TEMPLATE with a non-negative "subscript" that is not in NAME-SET.
;; E.g. 
;; 
;; > (name-not-in 'a '(b c d))
;; a_0
;; 
;; > (name-not-in 'a '(a b c d))
;; a_0
;; 
;; > (name-not-in 'a '(b c a_0 d))
;; a_1
;; 
;; > (name-not-in 'a '(a_0 a_4 a_2 b a_1 c d))
;; a_3
;; 
;; > (name-not-in 'a_0 '(a_0 a_4 a_2 b a_1 c d))
;; a_3
;; 
;; > (name-not-in 'a_3 '(a_0 a_4 a_2 b a_1 c d))
;; a_3

;; The implementation below uses several functions --
;; pre-defined functions from the Scheme Standard Library: 
;;    string->symbol, string-append, symbol->string, 
;;    number->string, substring, string-length, string-ref, 
;;    char>=?, and char<=?,
;; help functions defined in this file:
;;    name-search, root, subscript, root-loop, and digit?,
;; help functions defined in the file "set-list-nodups.scm":
;;    set-member?

(define name-not-in
  (lambda (template name-set)
    (name-search (root template) 0 name-set)))

(define name-search 
  (lambda (symbol int name-set)
    (let ((subscripted (subscript symbol int)))
      (if (not (set-member? subscripted name-set))
        subscripted
        (name-search symbol (+ int 1) name-set)))))

(define subscript
  (lambda (symbol int) 
    (string->symbol 
     (string-append
      (symbol->string symbol)
      "_"
      (number->string int)))))

;; Returns the "root" of a name -- i.e., the part before
;; any final underscore followed by a number.
(define root
  (lambda (symbol)
    (let ((str (symbol->string symbol)))
      (string->symbol 
       (substring str
		  0
		  (root-loop str (- (string-length str) 1) #f))))))

;; Returns index of final underscore in string if it is followed
;; by a non-empty number of digits. In all other cases, returns
;; length of string. 
(define root-loop
  (lambda (str index seen-digit?)
    (let ((char (string-ref str index)))
      (cond ((= index 0) (string-length str))
            ((char=? char #\_)
	     (if seen-digit? index (string-length str)))
            ((digit? char)
	     (root-loop str (- index 1) #t))
            (else (string-length str))))))

      
(define digit? 
  (lambda (char)
    (or (char>=? char #\0) (char<=? char #\9))))
