;;; SUBST.SCM for INTEX+BIND programs

;; (subst <env> <exp>) returns a copy of expression <exp> 
;; in which every variable reference to a name bound in <env>
;; is replaced by the expression bound to the name. The entity
;; bound to the name should be an initlit node, *not* an integer. 
(define subst
  (lambda (env exp) 
    ;; rename binding and bound occurrences in EXP to obtain RENAMED-EXP,
    ;; then carry out case-analysis of RENAMED-EXP using LOCAL-SUBST,
    ;; which is defined recursively and takes one argument only.
    (letrec ((renamed-exp (rename exp (find-reserved-var-names env exp)))
	     (local-subst (lambda (exp)
                 (cond ((literal? exp) exp)
		       ((varref? exp) 
			(let ((probe (env-lookup (varref-name exp) env)))
			  (if (unbound? probe) exp probe)))
		       ((binapp? exp)
			(make-binapp (binapp-rator exp)
				     (local-subst (binapp-rand1 exp))
				     (local-subst (binapp-rand2 exp))))
		       ((bind? exp)
			(let ((name (bind-name exp))
			      (defn (bind-defn exp))
			      (body (bind-body exp)))
			  (make-bind name 
				     (local-subst defn)
				     (local-subst body))))
		       (else (error "SUBST: Unrecognized expression -- " exp))
          ))))
      (local-subst renamed-exp)
      )))

;; FIND-RESERVED-VAR-NAMES is an auxiliary function of two
;; arguments, an environment ENV and an expression EXP.
;; (FIND-RESERVED-VAR-NAMES ENV EXP) returns the set consisting of:
;;    * all variable names occurring free in EXP,
;;    * all variable names occurring in ENV-KEYS of ENV
;;      which are not shadowed,
;;    * all variable names occurring free in ENV-VALUES of ENV
;;      which are not shadowed.
;; A key/value pair is "shadowed" by another key1/value1 pair in ENV 
;; if key = key1 and the pair key1/value1 occurs before, i.e., to the 
;; left of, the pair key/value in ENV.

(define find-reserved-var-names
  (lambda (env exp)
    (set-union (free-vars exp)
	       (set-union (env-keys env)
			  (foldr set-union ()
				 (map free-vars (env-values env)))))))

;; Auxiliary functions for substitution 
(define subst1
  (lambda (new old exp)
    (subst (env-make (list old) (list new)) exp)))

(define subst*
  (lambda (news olds exp)
    (subst (env-make olds news) exp)))
